
import json
from abc import abstractmethod

from .utils import Storage

class ConfigBase(object):
    def __init__(self):
        self._subscribers = []

    def on_change(self, handler):
        self._subscribers.append(handler)
        return self

    def notify(self, new, old):
        for handler in self._subscribers:
            handler(new, old)
        return self

class ConfigValue(ConfigBase):
    def __init__(self, value):
        ConfigBase.__init__(self)
        self.value = value

    def __repr__(self):
        return str(self.value)

class ConfigNode(dict, ConfigBase):
    def __init__(self, *args, **kwargs):
        ConfigBase.__init__(self)
        dict.__init__(self, *args, **kwargs)

    def sub_values(self):
        subs = dict()
        for name in self:
            val = self[name]
            if type(val) is ConfigNode:
                subs[name] = val.sub_values()
            else:
                subs[name] = val.value
        return subs

class Config(object):

    def __init__(self, logger):
        self.data = Storage()
        self.logger = logger
        self.raw_data = Storage()

    def raw(self):
        return self.raw_data

    def load(self, data):
        self.raw_data = data
        return self.parse()

    def reload(self):
        self.old_data = self.data
        if not self.parse():
            return False

        self.logger.info('Reloading the config')

        def check_eq(new_node, old_node):

            new_node._subscribers = old_node._subscribers

            res = False

            if type(new_node) is ConfigNode and type(old_node) is ConfigNode:

                # common nodes in new_node and new_node
                common = []
                # get new nodes in new_node dict
                for name in new_node:
                    if name in old_node:
                        common.append(name)
                    else:
                        new_node.notify(name, None) #...
                for name in old_node:
                    if name not in new_node:
                        old_node.notify(None, name)

                for name in common:
                    res = res or check_eq(new_node[name], old_node[name])

                if res:
                    new_node.notify(new_node, old_node)

            elif type(new_node.value) is list and type(old_node.value) is list:
                if len(new_node.value) != len(old_node.value):
                    new_node.notify(new_node, old_node) #jeeeeeeeeeeeeeej
                    res = True
            else:
                if new_node.value != old_node.value:
                    new_node.notify(new_node, old_node) #jeeeeeeeeeeeeeej
                    res = True

            return res

        check_eq(self.data, self.old_data)

    def parse(self):

        data = ConfigNode(self.raw_data)

        def iter(node):
            for name in node:
                val = node[name]
                if type(val) is dict:
                    node[name] = ConfigNode(node[name])
                    iter(node[name])
                else:
                    node[name] = ConfigValue(node[name])

        iter(data)
        self.data = data

        return True



from threading import Thread, Event

class Timer(Thread):
    def __init__(self, interval, callback):
        super(Timer, self).__init__()

        self.interval_event = Event()
        self.main_event = Event()
        self.interval = interval
        self.callback = callback

        self.daemon = True # stop if the program exits

        # start the thread (not the timer...)
        super(Timer, self).start()

    def run(self):
        while self.main_event.wait():
            while not self.interval_event.wait(self.interval):
                th = Thread(target=self.callback)
                th.setDaemon(True)
                th.start()

    def start(self, interval = None):
        if interval is not None:
            self.interval = interval
        self.interval_event.clear()
        self.main_event.set()

    def is_running(self):
        return not self.interval_event.is_set()

    def stop(self):
        self.interval_event.set()




import importlib
import os
import urllib2, urllib
import re
import sys
import json

def ucfirst(str):
    return str[0].upper() + str[1:]

def split_uppercase(str):
    return re.findall('[A-Z][^A-Z]*', str)

def xml_element_to_storage(element):

    res = Storage()
    for item in element:
        res[item.tag] = item.text

    return res

def extend_path(root_dir, paths):
    new_paths = []
    for path in paths:
        pp = [root_dir]
        if type(path) is list:
            pp.extend(path)
        else:
            pp.append(path)

        new_path = os.path.join(*pp)
        new_paths.append(new_path)

    sys.path.extend(new_paths)

class Platforms():
    folders = dict(
        win32 = 'win',
        linux2 = 'linux',
        linux = 'linux',
    )

    def get_folder(self):
        platform = sys.platform
        if platform not in self.folders:
            raise Exception('Unknown platform %s' % platform)

        return self.folders[platform]

    def import_class(self, name):
        folder = self.get_folder()
        file_name = '_'.join(split_uppercase(name)).lower()
        module = importlib.import_module('%s.%s' % (folder, file_name))
        return getattr(module, name)

class Storage(dict):

    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, value):
        self[key] = value

    def __hasattr__(self, key):
        return key in self


class FileReader(object):
    def read_all(self, file_name):
        with open(file_name) as f:
            content = f.read()
        return content


class SimpleResponse(Storage):
    def __init__(self, code, message = ''):
        self.code = code
        self.message = message

    @staticmethod
    def from_string(response):
        try:
            data = json.loads(response)
            return SimpleResponse(data['code'], data['message'])
        except:
            raise

    def __str__(self):
        return "Code: %r, message: %s" % (self.code, self.message)

class HTTPResponse(object):
    def __init__(self, code, content, headers = {}):
        self.code = code
        self.content = content
        self.headers = headers

    def __str__(self):
        return "code: %d, headers: %s, content: %s" % (self.code, self.headers, self.content)

class URLLoader(object):
    def load(self, url, data = None, headers = {}):

        if data:
            data = urllib.urlencode(data)

        req = urllib2.Request(url, data, headers)

        try:
            response = urllib2.urlopen(req)
            return HTTPResponse(response.getcode(), response.read(), response.info().dict)
        except urllib2.URLError, e:
            return HTTPResponse(e.code, e.read(), e.info().dict)



# -*- coding: utf-8 -*-

import os
import sys
import signal
import time
import websocket

from threading import Thread

from logger import get_logger
from config import Config
from utils import FileReader
from message import Message
from tools.timer import Timer
import editors

class AbsintheClient():

    def __init__(self, config, logger):
        self.config = config.raw()
        self.logger = logger

    def start(self):
        self.create_hosts(self.config['hosts'])
        self.create_editors(self.config['editors'])
        self.create_sessions(self.config['sessions'])

        for name in self.hosts:
            host = self.hosts[name]
            if len(host.on_connect_callbacks):
                host.connect()

        # a very simple blocking mecha...
        while True:
            try:
                sys.stdin.read(1)
            except KeyboardInterrupt as k:
                self.stop()
                break

    def create_sessions(self, sessions):
        self.logger.debug('Create sessions')
        self.sessions = {}
        for session in sessions:
            if session['enabled'] != True:
                continue
            name = session['name']
            if session['host'] not in self.hosts:
                self.logger.error('Undefined host %s in session %s. Ignoring session.' % (session['host'], name))
                continue

            if session['editor'] not in self.editors:
                self.logger.error('Undefined editor %s in session %s. Ignoring session.' % (session['editor'], name))
                continue

            self.sessions[name] = Session(name, self.hosts[session['host']], self.editors[session['editor']], session['path'], session['remote-path'], self.logger)

    def create_editors(self, editors_defs):
        self.logger.debug('Create editors')
        self.editors = {}
        for name in editors_defs:
            if not hasattr(editors, name):
                self.logger.error('Undefined editor: %s. Skipping.' % name)
                continue
            TypedEditor = getattr(editors, name)
            self.editors[name] = TypedEditor(name, editors_defs[name], self.logger)

    def create_hosts(self, hosts):
        self.logger.debug('Create hosts')
        self.hosts = {}
        for name in hosts:
            addr = hosts[name]
            self.hosts[name] = Host(name, addr['host'], addr['port'], self.logger)

    def stop(self):
        pass

class Host():
    def __init__(self, name, host, port, logger):
        self.msg_queue = []
        self.logger = logger
        self.name = name
        self.host = "ws://%s:%d" % (host, port)

        self.sessions = {}
        self.connected = False
        self.on_connect_callbacks = []

        self.reconnect_timer = Timer(10, self.connect)

    def connect(self):
        self.logger.debug('Try to connect to %s - %s' % (self.name, self.host))
        self.socket = websocket.WebSocketApp(self.host, on_message = self.on_message, on_error = self.on_error, on_close = self.on_close)
        self.socket.on_open = self.on_open

        th = Thread(target=self.socket.run_forever)
        th.setDaemon(True)
        th.start()

    def on_connect(self, callback):
        self.on_connect_callbacks.append(callback)

    def send(self, message):
        if not self.is_connected():
            self.logger.error('Try to send message, but no open socket. Name: %s, Message: %s' % (self.name, message))
            self.msg_queue.append(message)
            return

        self.logger.debug('Send message to %s: %s' % (self.host, message))

        self.socket.send(str(message))

    def is_connected(self):
        return self.connected

    def subscribe(self, name, session):
        if name not in self.sessions:
            self.sessions[name] = []

        self.sessions[name].append(session)

    def on_open(self, ws):
        self.connected = True
        self.reconnect_timer.stop()
        self.logger.info('Connected to %s - %s' % (self.name, self.host))

        for callback in self.on_connect_callbacks:
            callback()

        for msg in self.msg_queue:
            self.send(msg)
        self.msg_queue = []

    def on_message(self, ws, message):
        try:
            msg = Message.from_str(message)
        except Exception as e:
            self.logger.error('Malformed message received via websocket: %s, %s' % (e, message))
            return

        self.logger.debug('Message received %s' % msg)

        if msg.name not in self.sessions:
            self.logger.error('Unknown session %s' % msg.name)
            return

        for session in self.sessions[msg.name]:
            if hasattr(session, msg.command):
                func = getattr(session, msg.command)
                func(**msg.arguments)
            else:
                self.logger.error('Unknown command %s for session %s' % (msg.command, session.name))

    def on_error(self, ws, error):
        self.logger.error('Error occured on websocket: %s' % s)

    def on_close(self, reason):
        if self.is_connected():
            self.logger.error('%s socket closed, try to reconnect...' % self.name)
            self.connected = False
        self.reconnect_timer.start()

class Session():
    def __init__(self, name, host, editor, path, remote_path, logger):
        self.name = name
        self.host = host
        self.editor = editor
        self.path = path
        self.logger = logger
        self.remote_path = remote_path

        host.subscribe(name, self)
        host.on_connect(self.on_host_connected)

    def on_host_connected(self):
        msg = Message(self.name, 'session_start', dict(remote_path = self.remote_path))
        self.host.send(msg)

    def set_focus(self):
        self.editor.set_focus()

    def open_content(self, content):
        pass

    def open_file(self, filename, line = 1):

        full_path = os.path.join(self.path, *filename)

        self.logger.debug('Open file: %s' % full_path)

        self.editor.open_file(full_path, line)




import sys, os, time, atexit
from signal import SIGTERM

class Daemon(object):
    """
    A generic daemon class.

    Usage: subclass the Daemon class and override the run() method
    """
    def __init__(self, pidfile, logger):
        self.logger = logger
        self.pidfile = pidfile


    def init(self):
        """
        Called before the daemonize, so useful some init before the fork, eg print the user config parse errors, etc...
        return False to stop the daemonize!
        """
        return (True, 'Ok')

    def daemonize(self):
        """
        do the UNIX double-fork magic, see Stevens' "Advanced
        Programming in the UNIX Environment" for details (ISBN 0201563177)
        http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC16
        """
        try:
            pid = os.fork()
            if pid > 0:
                # exit first parent
                sys.exit(0)
        except OSError, e:
            self.logger.error("Fork #1 failed: %d (%s)\n" % (e.errno, e.strerror))
            sys.exit(1)

        # decouple from parent environment
        os.chdir("/")
        os.setsid()
        os.umask(0)

        # do second fork
        try:
            pid = os.fork()
            if pid > 0:
                # exit from second parent
                sys.exit(0)
        except OSError, e:
            self.logger.error("Fork #2 failed: %d (%s)\n" % (e.errno, e.strerror))
            sys.exit(1)

        # write pidfile
        atexit.register(self.delpid)
        pid = str(os.getpid())
        file(self.pidfile,'w+').write("%s\n" % pid)

    def delpid(self):
        os.remove(self.pidfile)

    def get_pid(self):
        pid = None

        try:
            with open(self.pidfile, 'r') as f:
                try:
                    pid = int(f.read().strip())
                except TypeError as e:
                    pid = None
        except IOError:
            pid = None

        return pid

    def is_running(self):

        pid = self.get_pid()

        if pid is None:
            return False

        try:
            os.kill(pid, 0)
        except OSError:
            return False
        else:
            return True

    def status(self):
        return 'Daemon is ' + ('running' if self.is_running() else 'not running')

    def start(self):
        """
        Start the daemon
        """

        if self.is_running():
            msg = "Daemon already running (pidfile:%s)" % self.pidfile
            self.logger.error(msg)
            return msg

        initres = self.init()

        if not initres[0]:
            return initres[1]

        # Start the daemon
        self.daemonize()
        return self.run()

    def stop(self):
        """
        Stop the daemon
        """
        # Get the pid from the pidfile
        pid = self.get_pid()

        if not pid:
            message = "Pidfile %s does not exist. Daemon not running?" % self.pidfile
            self.logger.error(message)
            return message # not an error in a restart

        # Try killing the daemon process
        try:
            while 1:
                os.kill(pid, SIGTERM)
                time.sleep(0.1)
        except OSError, err:
            err = str(err)
            if err.find("No such process") > 0:
                if os.path.exists(self.pidfile):
                    os.remove(self.pidfile)
            else:
                self.logger.error(err)
                sys.exit(1)

        return 'Daemon is stopped'

    def restart(self):
        """
        Restart the daemon
        """
        self.stop()
        self.start()

    def run(self):
        """
        You should override this method when you subclass Daemon. It will be called after the process has been
        daemonized by start() or restart().
        """



import subprocess
from abc import abstractmethod

from window_manager import WindowManager

class EditorBase(object):
    def __init__(self, name, bin_path, logger):
        self.name = name
        self.bin_path = bin_path
        self.logger = logger

    @abstractmethod
    def open_content(self, content):
        pass

    @abstractmethod
    def set_focus(self):
        pass

    @abstractmethod
    def get_command_pattern(self):
        return []

    def open_file(self, filename, line):
        command = []
        data = dict(filename = filename, line = line)
        for pattern in self.get_command_pattern():
            command.append(pattern % data)

        self.logger.debug('Call command to open file: %s' % command)

        try:
            subprocess.call(command)
        except Exception as e:
            self.logger.error('Editor command failed: %s' % e)


class Notepadpp(EditorBase):
    def __init__(self, *args):
        super(Notepadpp, self).__init__(*args)
        self.winmgr = WindowManager()

    def get_command_pattern(self):
        return [self.bin_path, "%(filename)s", "-n%(line)d"]

    def open_content(self, content):
        pass

    def set_focus(self):
        self.winmgr.find_window_regex(".*Notepad\+\+")
        self.winmgr.set_foreground()



import json

class Message(object):
    def __init__(self, name, command, arguments = dict()):
        self.name = name
        self.command = command
        self.arguments = arguments

    @staticmethod
    def from_str(message):
        try:
            data = json.loads(message)
        except:
            raise

        try:
            name = data['name']
            command = data['command']
            arguments = data['arguments']
        except:
            raise

        return Message(name, command, arguments)

    def __str__(self):
        return json.dumps(dict(name = self.name, command = self.command, arguments = self.arguments))




import threading
import pyjsonrpc
import json
import os
from geventwebsocket import WebSocketServer, WebSocketApplication, Resource

from absinthe.tools.commands import CommandRequestHandler, external_jsonrpc_command
from absinthe.message import Message
from absinthe.tools.utils import SimpleResponse
from absinthe.tools.remote_process_base import RemoteProcessBase

# gevent socket in thread, need to patch...
from gevent import monkey
monkey.patch_all()

class Client(WebSocketApplication):
    def __init__(self, ws, manager):
        WebSocketApplication.__init__(self, ws)
        self.manager = manager
        self.address = '%(REMOTE_ADDR)s:%(REMOTE_PORT)s' % ws.environ

    def send(self, message):
        self.ws.send(str(message))

    def on_open(self):
        pass

    def on_message(self, message):
        if message is None:
            return
        self.manager.on_message(message, self)

    def on_close(self, reason):
        self.manager.on_close(self)

class Session(object):
    def __init__(self, name, client, path):
        self.name = name
        self.client = client
        self.path = path

class PathHandler(object):
    def __init__(self, name, path):
        self.name = name
        self.path = path

    def parse(self, filename):
        if filename.find(self.path) != 0:
            raise Exception('Path mismatch')

        fn = filename[len(self.path):]

        return fn.split(os.sep)

class ClientManager(object):
    def __init__(self, session_manager):
        self.session_manager = session_manager

    def __call__(self, ws):
        return Client(ws, self.session_manager)

class SessionManager(object):
    def __init__(self, logger):
        self.logger = logger
        self.sessions = {}
        self.paths = {}

    def register_path(self, path):
        self.paths[path.name] = path

    def find_sessions(self, path_name):
        sessions = []
        for name in self.sessions:
            for session in self.sessions[name]:
                if session.path.name == path_name:
                    sessions.append(session)
        return sessions

    @external_jsonrpc_command
    def set_focus(self, path_name):
        sessions = self.find_sessions(path_name)

        for session in sessions:
            session.client.send(Message(session.name, 'set_focus'))

        return SimpleResponse(True)

    @external_jsonrpc_command
    def open_file(self, path_name, filename, line):
        self.logger.debug('Open file %s in %s' % (path_name, filename))

        msgs = []

        try:
            sessions = self.find_sessions(path_name)

            if len(sessions) == 0:
                msg = 'There is no client for this path %s' % path_name
                self.logger.warning(msg)
                return SimpleResponse(False, [msg])

            msgs.append('Session found: %s' % path_name)

            for session in sessions:
                file_parts = session.path.parse(filename)
                session.client.send(Message(session.name, 'open_file', dict(filename = file_parts, line = line)))
                msgs.append('File open request sent to %s' % session.client.address)

            for msg in msgs:
                self.logger.debug(msg)

        except Exception as e:
            self.logger.exception(e);
            msgs.append(e);

        return SimpleResponse(True, msgs)

    def on_message(self, message, client):
        try:
            msg = Message.from_str(message)
        except Exception as e:
            self.logger.error('Malformed message received via websocket: %s, %s' % (e, message))
            return

        if hasattr(self, msg.command):
           func = getattr(self, msg.command)
           func(msg.name, client, **msg.arguments)
        else:
            self.logger.warning('Undefined command received: %s' % msg.command)

    def on_close(self, client):
        for name in self.sessions:
            for session in self.sessions[name]:
                if session.client == client:
                    self.sessions[name].remove(session)
                    self.logger.info('Session close: %s from %s' % (name, client.address))

    def session_start(self, name, client, remote_path):
        self.logger.info('Session start: %s from %s' % (name, client.address))
        session = Session(name, client, self.paths[remote_path])
        if name not in self.sessions:
            self.sessions[name] = []

        self.sessions[name].append(session)

class AbsintheServer(RemoteProcessBase):

    def __init__(self, config, logger):
        self.logger = logger
        self.config = config
        self.session_manager = SessionManager(self.logger)
        self.client_manager = ClientManager(self.session_manager)

    @external_jsonrpc_command
    def init(self):
        for name in self.config.data['paths']:
            self.session_manager.register_path(PathHandler(name, self.config.data['paths'][name].value))

        server_address = self.config.data['agent_server']
        self.server = WebSocketServer((server_address['host'].value, server_address['port'].value), Resource({'/': self.client_manager}))

        th = threading.Thread(target=self.server.serve_forever)
        th.setDaemon(True)
        th.start()

        self.logger.debug('init')

        return SimpleResponse(True)

    # Initialize the command server to receive IPC commands.
    def start_command_server(self):
        try:
            command_server_address = self.config.data['command_server']
            self.command_server = pyjsonrpc.ThreadingHttpServer(
                server_address = (command_server_address['host'].value, command_server_address['port'].value),
                RequestHandlerClass = CommandRequestHandler
            )
        except Exception as e:
            self.logger.error('Exception occured during the command server initalization: ' + str(e) + traceback.format_exc())
            return

        CommandRequestHandler.logger = self.logger
        CommandRequestHandler.externals.extend([self, self.session_manager])

        self.logger.debug('command server starting...')

        self.command_server.serve_forever()




import win32serviceutil
import win32service
import win32event
import win32api

class Service(win32serviceutil.ServiceFramework):
    _svc_name_ = '_unNamed'
    _svc_display_name_ = '_Service Template'

    class Logger():
        def error(self, msg):
            import servicemanager
            servicemanager.LogErrorMsg(str(msg))

        def info(self, msg):
            import servicemanager
            servicemanager.LogInfoMsg(str(msg))

        def debug(self, msg):
            import servicemanager
            servicemanager.LogInfoMsg(str(msg))

        def warning(self, msg):
            import servicemanager
            servicemanager.LogWarningMsg(str(msg))

    def __init__(self, *args):
        win32serviceutil.ServiceFramework.__init__(self, *args)
        self.log('init')
        self.stop_event = win32event.CreateEvent(None, 0, 0, None)
        self.logger = Service.Logger()

    def log(self, msg):
        import servicemanager
        servicemanager.LogInfoMsg(str(msg))

    def sleep(self, sec):
        win32api.Sleep(sec*1000, True)

    def SvcDoRun(self):
        self.ReportServiceStatus(win32service.SERVICE_START_PENDING)
        try:
            self.ReportServiceStatus(win32service.SERVICE_RUNNING)
            self.log('start')
            self.start()
            self.log('wait')
            win32event.WaitForSingleObject(self.stop_event, win32event.INFINITE)
            self.log('done')
        except Exception, x:
            self.log('Exception : %s' % x)
            self.SvcStop()

    def SvcStop(self):
        self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)
        self.log('stopping')
        self.stop()
        self.log('stopped')
        win32event.SetEvent(self.stop_event)
        self.ReportServiceStatus(win32service.SERVICE_STOPPED)

    # to be overridden
    def start(self):
        pass

    # to be overridden
    def stop(self):
        pass



import re

class WindowManager(object):

    def __init__ (self):
        pass

    def find_window(self, class_name, window_name = None):
        pass

    def find_window_regex(self, regex):
        pass

    def set_foreground(self):
        """put the window in the foreground"""
        pass



import os
import shutil

from .tools.utils import Storage

dirs = Storage(
    package_path = os.path.abspath(os.path.dirname(__file__)),
    data_dir = os.path.join(os.path.expanduser('~'), '.absinthe'),
)

dirs.update(Storage(
    server_listener = os.path.join(dirs.package_path, 'server_listen.py'),
    config_file = os.path.join(dirs.data_dir, 'config.json'),
))

def init_data_dir():
    if not os.path.isdir(dirs.data_dir):
        os.mkdir(dirs.data_dir)

    if not os.path.isfile(dirs.config_file):
        shutil.copyfile(os.path.join(dirs.package_path, 'config_example.json'), dirs.config_file)



import logging
import logging.config

import json
import sys

class StdWriter(object):
    def __init__(self, logger, level):
        self.logger = logger
        self.level = level

    def write(self, text):
        self.logger.log(self.level, text)


def get_logger(config_data, name):
    logging.config.dictConfig(config_data)

    logger = logging.getLogger(name)

    sys.stderr = StdWriter(logger, logging.ERROR)
    sys.stdout = StdWriter(logger, logging.WARNING)

    return logger



import pyjsonrpc

from utils import SimpleResponse, Storage

class commandline(object):
    def __init__(self, help, arguments = dict()):
        self.help = help
        self.arguments = arguments

    def __call__(self, func):
        func.commandline = Storage(
            help = self.help,
            arguments = self.arguments
        )

        return func

class CommandRequestHandler(pyjsonrpc.HttpRequestHandler):
    logger = None
    externals = []

    def log_message(self, format, *args):
        # called from parent class...
        self.logger.debug("HTTP request - %s - %s" % (self.client_address[0], format % args))

    @commandline('Ping')
    @pyjsonrpc.rpcmethod
    def ping(self):
        return SimpleResponse(True, 'pong')

    """
    Example of json rpc method with parameters when should be callable from command line:

    @commandline('paramed test method', dict(teve=dict(help='teveee', type=int), alma=dict(help = 'set the alma', type = int)))
    @pyjsonrpc.rpcmethod
    def paramed(self, alma, teve):
        self.logger.info('paramed ' + str(alma) + str(teve))
        return SimpleResponse(True, alma*teve)
    """


"""
    This decorator function make an external class's methods to able to called from json rpc client
    For examples see Plutonium class (the order of the decorators are important!)
"""
def external_jsonrpc_command(orig_func):

    def runner(*args, **kwargs):

        ext_obj = None
        for external in CommandRequestHandler.externals:
            if hasattr(external, orig_func.func_name):
               ext_obj = external
               break

        if ext_obj is None:
            return

        # get the original function in the original external class
        func = getattr(ext_obj, orig_func.func_name)

        # need to remove the CommandRequestHandler instance from the beginning of args tuple
        args_list = list(args)
        args_list.pop(0)
        args_tuple = tuple(args_list)

        return func(*args_tuple, **kwargs)

    # add the runner function to CommandRequestHandler by the decorated function name
    setattr(CommandRequestHandler, orig_func.func_name, runner)

    # call the orig decorator
    pyjsonrpc.rpcmethod(runner)

    # need to copy commandline params if exists
    if hasattr(orig_func, 'commandline'):
        runner.commandline = orig_func.commandline

    return orig_func



import win32gui, win32com.client
import re
import threading
import pythoncom

class WindowManager(object):
    """Encapsulates some calls to the winapi for window management"""
    def __init__ (self):
        """Constructor"""
        self._handle = None

    def find_window(self, class_name, window_name = None):
        """find a window by its class_name"""
        self._handle = win32gui.FindWindow(class_name, window_name)

    def _window_enum_callback(self, hwnd, regex):
        '''Pass to win32gui.EnumWindows() to check all the opened windows'''
        if re.match(regex, str(win32gui.GetWindowText(hwnd))) != None:
            self._handle = hwnd

    def find_window_regex(self, regex):
        self._handle = None
        win32gui.EnumWindows(self._window_enum_callback, regex)

    def set_foreground(self):
        """put the window in the foreground"""

        try:
            # "this is not the main thread" problem (CoInitialize has not been called)
            if threading.currentThread().getName() <> 'MainThread':
                pythoncom.CoInitialize()

            # "this is not the active window" problem (pywintypes.error: (0, 'SetForegroundWindow', 'No error message is available'))
            shell = win32com.client.Dispatch("WScript.Shell")
            shell.SendKeys('^')

            win32gui.SetForegroundWindow(self._handle)
        except Exception as e:
            print e




import time

class Service(object):

    # to be overridden
    def start(self):
        pass

    # to be overridden
    def stop(self):
        pass

    def log(self, msg):
        pass

    def sleep(self, sec):
        time.sleep(sec)





class ServiceManager(object):

    def __init__(self, cls, name, display_name = None):
        self.cls = cls

    def start(self):
        service = self.cls()
        service.start()

    def stop(self):
        pass

    def remove(self):
        pass

    def install(self, stay_alive=True):
        pass



from setuptools import setup, find_packages

setup(
    name = "Absinthe",
    version = '1.0.0',
    description = "A tiny server(-client) app to open sshfs connected files over ssh in any local editor",
    author = 'Lajos Santa',
    author_email = 'santa.lajos@coldline.hu',
    url = 'https://github.com/voidpp/absinthe',
    license = 'MIT',
    install_requires = [
        "gevent-websocket==0.9.3",
        "python-jsonrpc==0.7.12",
    ],
    packages = find_packages(),
    include_package_data = True,
    scripts = [
        'bin/absinthe',
        'bin/abs',
    ],
)



import os
import sys
import json

from absinthe import dirs
from absinthe.tools.logger import get_logger
from absinthe.tools.config import Config
from absinthe.absinthe_server import AbsintheServer

config_data = {}
with open(dirs.config_file) as f:
    config_data = json.load(f)

# init logger
logger = get_logger(config_data['logger'], 'server')
logger.info("Welcome to the lofasz. (pid: %d)" % os.getpid())

# init config
try:
    config = Config(logger)
    config.load(config_data['server'])
    logger.info('Config has been successfully loaded')
except Exception as e:
    logger.error('Exception occured during config parsing: ' + str(e))
    sys.exit(1)

server = AbsintheServer(config, logger)
server.start_command_server()




import time
import urllib2
import pyjsonrpc

from utils import Storage, SimpleResponse
from background_process_handler import BackgroundProcessHandler

# used by the init script to manage the remote process, eg start, stop, create the IPC interface
class RemoteProcessManager(BackgroundProcessHandler):

    def __init__(self, command, control_port, pid_file, logger):
        super(RemoteProcessManager, self).__init__(command, pid_file, logger)
        self.control_port = control_port

    def get_rpc_client(self):
        return pyjsonrpc.HttpClient(url = "http://localhost:%d/jsonrpc" % self.control_port)

    def start(self):
        result = super(RemoteProcessManager, self).start()

        if result.code is False:
            return result

        rpc_client = self.get_rpc_client()

        """
            The BackgroundProcessHandler.start function executes the remote script, and returns immediately.
            But the command server not available yet, so we need to wait for it.
        """
        attempts = 50

        while True:
            try:
                rpc_client.ping()
                break
            except urllib2.URLError as e:
                attempts = attempts - 1
                if attempts > 0:
                    time.sleep(0.1)
                else:
                    break

        if not attempts:
            # if the remote process unwilling to communicate, needs to stop it!
            self.stop()
            return SimpleResponse(False, 'Initialize has been failed')

        self.logger.debug('send init')

        initres = rpc_client.init()

        self.logger.debug(initres)

        init_res = Storage(initres)

        self.logger.debug('Init remote process: ' + str(init_res))

        if init_res.code:
            result.message = result.message + ' and initialized'

        return result

    def control(self, name, args = {}):

        self.logger.debug("Send rpc command '%s' with args: %s" % (name, args))

        rpc_client = self.get_rpc_client()
        func = getattr(rpc_client, name)

        response = func(**args)

        return SimpleResponse(**dict(response))




import os
import time
import subprocess
from signal import SIGTERM

from utils import SimpleResponse

class BackgroundProcessHandler(object):
    def __init__(self, command, pid_file, logger):
        self.pid_file = pid_file
        self.logger = logger
        self.command = command

    def start(self):
        if self.is_running():
            return SimpleResponse(False, 'Daemon is already running. Pid: %d' % self.get_pid())

        pid = subprocess.Popen(self.command).pid

        file(self.pid_file,'w+').write("%s\n" % pid)

        return SimpleResponse(True, 'Started (pid: %s)' % pid)

    def get_pid(self):
        pid = None

        try:
            with open(self.pid_file, 'r') as f:
                try:
                    pid = int(f.read().strip())
                except TypeError as e:
                    pid = None
        except IOError:
            pid = None

        return pid

    def is_running(self):
        pid = self.get_pid()

        if pid is None:
            return False

        try:
            os.kill(pid, 0)
        except OSError:
            return False
        else:
            return True

    def status(self):
        return SimpleResponse(True, 'Daemon is ' + ('running' if self.is_running() else 'not running'))

    def stop(self):
        # Get the pid from the pidfile
        pid = self.get_pid()

        if not pid:
            message = "Pidfile %s does not exist" % self.pid_file
            self.logger.error(message)
            return SimpleResponse(False, 'Daemon is not running')

        # Try killing the daemon process
        try:
            while 1:
                os.kill(pid, SIGTERM)
                time.sleep(0.1)
        except OSError, err:
            err = str(err)
            if err.find("No such process") > 0:
                if os.path.exists(self.pid_file):
                    os.remove(self.pid_file)
            else:
                self.logger.error(err)
                sys.exit(1)

        return SimpleResponse(True, 'Daemon is stopped')

    def restart(self):
        self.stop()
        return self.start()




from os.path import splitext, abspath
from sys import modules

import win32serviceutil
import win32service
import win32api

class ServiceManager(object):

    def __init__(self, cls, name, display_name = None):
        '''
            cls : the class (derived from Service) that implement the Service
            name : Service name
            display_name : the name displayed in the service manager
        '''
        self.cls = cls
        cls._svc_name_ = name
        cls._svc_display_name_ = display_name or name

        try:
            module_path = modules[cls.__module__].__file__
        except AttributeError:
            # maybe py2exe went by
            from sys import executable
            module_path = executable

        module_file = splitext(abspath(module_path))[0]
        cls._svc_reg_class_ = '%s.%s' % (module_file, cls.__name__)

    def start(self):
        try:
            win32serviceutil.StartService(self.cls._svc_name_)
            return True
        except Exception as e:
            # the service is not installed
            if e[0] in [1060, 1056]:
                return False
            else:
                raise

    def stop(self):
        try:
            win32serviceutil.StopService(self.cls._svc_name_)
            return True
        except Exception as e:
            # the service is not running
            if e[0] in [1060, 1062]:
                return False
            else:
                raise

    def remove(self):
        self.stop()
        try:
            svc_mgr = win32service.OpenSCManager(None,None,win32service.SC_MANAGER_ALL_ACCESS)
            svc_handle = win32service.OpenService(svc_mgr, self.cls._svc_name_, win32service.SERVICE_ALL_ACCESS)
            win32service.DeleteService(svc_handle)
            return True
        except Exception as e:
            # the service is not installed
            if e[0] in [1060]:
                return False
            else:
                raise

    def install(self, stay_alive=True):
        '''
            stay_alive : Service will stop on logout if False
        '''
        if stay_alive: win32api.SetConsoleCtrlHandler(lambda x: True, True)
        try:
            win32serviceutil.InstallService(
                self.cls._svc_reg_class_,
                self.cls._svc_name_,
                self.cls._svc_display_name_,
                startType = win32service.SERVICE_AUTO_START
            )
            return True
        except Exception as e:
            # the service is already
            if e[0] in [1073]:
                return False
            else:
                raise
